/******************************************************************************

  The DUNE (see www.dune-project.org) module dune-spgrid provides a structured,
  parallel grid satisfying the dune-grid interface.

  Copyright (C) 2012 - 2014 Christoph Gersbacher
  Copyright (C) 2010 - 2013 Robert Klöfkorn
  Copyright (C) 2013        Tobias Malkmus
  Copyright (C) 2009 - 2014 Martin Nolte

  The dune-spgrid module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation, either version 2 of 
  the License, or (at your option) any later version.

  The dune-spgrid module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_SPGRID_GRID_HH
#define DUNE_SPGRID_GRID_HH

#include <dune/common/parallel/mpicollectivecommunication.hh>

#include <dune/geometry/genericgeometry/codimtable.hh>

#include <dune/grid/common/grid.hh>
#include <dune/grid/common/adaptcallback.hh>
#include <dune/grid/common/geometryreference.hh>
#include <dune/grid/utility/grapedataioformattypes.hh>

#include <dune/grid/spgrid/capabilities.hh>
#include <dune/grid/spgrid/entityseed.hh>
#include <dune/grid/spgrid/gridview.hh>
#include <dune/grid/spgrid/hierarchiciterator.hh>
#include <dune/grid/spgrid/idset.hh>
#include <dune/grid/spgrid/indexset.hh>
#include <dune/grid/spgrid/hindexset.hh>
#include <dune/grid/spgrid/fileio.hh>

namespace Dune
{

  // External Forward Declarations
  // -----------------------------

  template< class Grid >
  struct BackupRestoreFacility;



  // Internal Forward Declarations
  // -----------------------------

#if HAVE_MPI
  template< class ct, int dim, SPRefinementStrategy strategy = SPIsotropicRefinement, class Comm = MPI_Comm >
  class SPGrid;
#else
  template< class ct, int dim, SPRefinementStrategy strategy = SPIsotropicRefinement, class Comm = No_Comm >
  class SPGrid;
#endif // #if !HAVE_MPI



  // SPGridFamily
  // ------------

  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  struct SPGridFamily
  {
    struct Traits
    {
      typedef SPGrid< ct, dim, strategy, Comm > Grid;

      typedef SPReferenceCubeContainer< ct, dim > ReferenceCubeContainer;
      typedef typename ReferenceCubeContainer::ReferenceCube ReferenceCube;
      typedef SPDomain< ct, dim > Domain;
      typedef SPMesh< dim > Mesh;
      typedef SPRefinement< dim, strategy > Refinement;
      typedef typename Refinement::Policy RefinementPolicy;

      typedef typename SPCommunicationTraits< Comm >::CollectiveCommunication CollectiveCommunication;

      typedef Dune::Intersection< const Grid, SPIntersection< const Grid > > LevelIntersection;
      typedef LevelIntersection LeafIntersection;

      typedef Dune::IntersectionIterator< const Grid, SPIntersectionIterator< const Grid >, SPIntersection< const Grid > > LevelIntersectionIterator;
      typedef LevelIntersectionIterator LeafIntersectionIterator;

      typedef Dune::EntityIterator< 0, const Grid, SPHierarchicIterator< const Grid > > HierarchicIterator;

      typedef SPIndexSet< const Grid > LevelIndexSet;
      typedef LevelIndexSet LeafIndexSet;

      typedef SPGlobalIdSet< const Grid > GlobalIdSet;
      typedef SPLocalIdSet< const Grid > LocalIdSet;
      typedef unsigned long GlobalIdType;
      typedef unsigned long LocalIdType;

      template< int codim >
      struct Codim
      {
        typedef typename ReferenceCubeContainer::template Codim< codim >::ReferenceCube ReferenceCube;

        typedef Dune::Entity< codim, dim, const Grid, SPEntity > Entity;

        typedef Dune::EntitySeed< const Grid, SPEntitySeed< codim, const Grid > > EntitySeed;

        typedef SPEntityPointer< codim, const Grid > EntityPointerImpl;
        typedef Dune::EntityPointer< const Grid, EntityPointerImpl > EntityPointer;

        typedef SPLocalGeometry< dim - codim, dim, const Grid > LocalGeometryImpl;

        typedef Dune::Geometry< dim - codim, dim, const Grid, SPGeometry > Geometry;
        typedef Dune::Geometry< dim - codim, dim, const Grid, LocalGeometryReference > LocalGeometry;

        template< PartitionIteratorType pitype >
        struct Partition
        {
          typedef Dune::EntityIterator< codim, const Grid, SPIterator< codim, pitype, const Grid > >
            LevelIterator;
          typedef Dune::EntityIterator< codim, const Grid, SPIterator< codim, pitype, const Grid > >
            LeafIterator;
        };

        typedef typename Partition< All_Partition >::LevelIterator LevelIterator;
        typedef typename Partition< All_Partition >::LeafIterator LeafIterator;
      };

      template< PartitionIteratorType pitype >
      struct Partition
      {
        typedef Dune::GridView< SPGridViewTraits< const Grid, pitype > > LevelGridView;
        typedef Dune::GridView< SPGridViewTraits< const Grid, pitype > > LeafGridView;
      };
    };
  };



  /** \class SPGrid
   *  \brief structured, parallel <a href="http://www.dune-project.org">%DUNE</a> grid
   *
   *  \tparam  ct        coordinate type (e.g., double)
   *  \tparam  dim       dimension of the grid
   *  \tparam  strategy  refinement strategy (default is \ref SPRefinementStrategy "SPIsotropicRefinement")
   *  \tparam  Comm      type of communicator (default depends on HAVE_MPI)
   */
  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  class SPGrid
  : public GridDefaultImplementation< dim, dim, ct, SPGridFamily< ct, dim, strategy, Comm > >
  {
    typedef SPGrid< ct, dim, strategy, Comm > This;
    typedef GridDefaultImplementation< dim, dim, ct, SPGridFamily< ct, dim, strategy, Comm > > Base;

    friend struct BackupRestoreFacility< This >;
    friend class SPIntersection< const This >;
    friend class SPGridLevel< This >;

  public:
    typedef SPGridFamily< ct, dim, strategy, Comm > GridFamily;

    typedef typename GridFamily::Traits Traits;

    typedef typename Traits::ReferenceCubeContainer ReferenceCubeContainer;
    typedef typename Traits::ReferenceCube ReferenceCube;
    typedef typename Traits::Domain Domain;
    typedef typename Traits::Mesh Mesh;
    typedef typename Traits::Refinement Refinement;
    typedef typename Traits::RefinementPolicy RefinementPolicy;

    typedef typename ReferenceCube::ctype ctype;

    static const int dimension = ReferenceCube::dimension;
    static const int dimensionworld = ReferenceCube::dimension;

    typedef typename ReferenceCube::GlobalVector GlobalVector;

    typedef typename Traits::GlobalIdSet GlobalIdSet;
    typedef typename Traits::LocalIdSet LocalIdSet;

    typedef typename Traits::CollectiveCommunication CollectiveCommunication;

    template< int codim >
    struct Codim
    : Base::template Codim< codim >
    {
      typedef typename Traits::template Codim< codim >::ReferenceCube ReferenceCube;

      typedef typename Traits::template Codim< codim >::LocalGeometryImpl LocalGeometryImpl;
    };

    template< PartitionIteratorType pitype >
    struct Partition
    : Base::template Partition< pitype >
    {};
    
    typedef typename Partition< All_Partition >::LevelGridView LevelGridView;
    typedef typename Partition< All_Partition >::LeafGridView LeafGridView;

    typedef typename LevelGridView::IndexSet LevelIndexSet;
    typedef typename LeafGridView::IndexSet LeafIndexSet;

    typedef SPHierarchyIndexSet< const This > HierarchicIndexSet;

    typedef SPGridLevel< This > GridLevel;

    typedef typename GridLevel::MultiIndex MultiIndex;
    static const int numDirections = GridLevel::numDirections;

  private:
    typedef typename GridLevel::PartitionList PartitionList;

    typedef typename LevelGridView::Traits::GridViewImp LevelGridViewImpl;
    typedef typename LeafGridView::Traits::GridViewImp LeafGridViewImpl;

  public:
    SPGrid ( const CollectiveCommunication &comm = SPCommunicationTraits< Comm >::defaultComm() );

    SPGrid ( const Domain &domain, const MultiIndex &cells,
             const CollectiveCommunication &comm = SPCommunicationTraits< Comm >::defaultComm() );

    SPGrid ( const Domain &domain, const MultiIndex &cells, const MultiIndex &overlap,
             const CollectiveCommunication &comm = SPCommunicationTraits< Comm >::defaultComm() );

    SPGrid ( const GlobalVector &a, const GlobalVector &b, const MultiIndex &cells,
             const CollectiveCommunication &comm = SPCommunicationTraits< Comm >::defaultComm() );

    SPGrid ( const GlobalVector &a, const GlobalVector &b, const MultiIndex &cells,
             const MultiIndex &overlap,
             const CollectiveCommunication &comm = SPCommunicationTraits< Comm >::defaultComm() );

    ~SPGrid ()
    {
      clear();
      for( int face = 0; face < ReferenceCube::numFaces; ++face )
        delete localFaceGeometry_[ face ];
    }

    using Base::getRealImplementation;

    const ReferenceCube &referenceCube () const
    {
      return refCubes_.get();
    }

    template< int codim >
    const typename Codim< codim >::ReferenceCube &referenceCube () const
    {
      return refCubes_.template get< codim >();
    }

    const Domain &domain () const { return domain_; }

    const MultiIndex &overlap () const { return overlap_; }

    int maxLevel () const
    {
      return leafLevel().level();
    }

    int size ( const int level, const int codim ) const
    {
      return levelGridView( level ).size( codim );
    }

    int size ( const int codim ) const
    {
      return leafGridView().size( codim );
    }

    int size ( const int level, const GeometryType &type ) const
    {
      return levelGridView( level ).size( type );
    }

    int size ( const GeometryType &type ) const
    {
      return leafGridView().size( type );
    }

    template< PartitionIteratorType pitype >
    typename Traits::template Partition< pitype >::LevelGridView
    levelGridView ( const int level ) const;

    template< PartitionIteratorType pitype >
    typename Traits::template Partition< pitype >::LeafGridView
    leafGridView () const;

    LevelGridView levelGridView ( const int level ) const;
    LeafGridView leafGridView () const;

    // for compatibility with dune-grid 2.3
    template< PartitionIteratorType pitype >
    typename Traits::template Partition< pitype >::LevelGridView
    levelView ( const int level ) const
    {
      return levelGridView< pitype >( level );
    }

    // for compatibility with dune-grid 2.3
    template< PartitionIteratorType pitype >
    typename Traits::template Partition< pitype >::LeafGridView
    leafView () const
    {
      return leafGridView< pitype >();
    }

    // for compatibility with dune-grid 2.3
    LevelGridView levelView ( const int level ) const { return levelGridView( level ); }
    // for compatibility with dune-grid 2.3
    LeafGridView leafView () const { return leafGridView(); }

    template< int codim, PartitionIteratorType pitype >
    typename Traits::template Codim< codim >::template Partition< pitype >::LevelIterator
    lbegin ( const int level, const unsigned int sweepDir = 0 ) const
    {
      const LevelGridView &view = levelGridView( level );
      return getRealImplementation( view ).template begin< codim, pitype >( sweepDir );
    }

    template< int codim, PartitionIteratorType pitype >
    typename Traits::template Codim< codim >::template Partition< pitype >::LevelIterator
    lend ( const int level, const unsigned int sweepDir = 0 ) const
    {
      const LevelGridView &view = levelGridView( level );
      return getRealImplementation( view ).template end< codim, pitype >( sweepDir );
    }

    template< int codim >
    typename Traits::template Codim< codim >::LevelIterator
    lbegin ( const int level, const unsigned int sweepDir = 0 ) const
    {
      const LevelGridView &view = levelGridView( level );
      return getRealImplementation( view ).template begin< codim >( sweepDir );
    }

    template< int codim >
    typename Traits::template Codim< codim >::LevelIterator
    lend ( const int level, const unsigned int sweepDir = 0 ) const
    {
      const LevelGridView &view = levelGridView( level );
      return getRealImplementation( view ).template end< codim >( sweepDir );
    }

    template< int codim, PartitionIteratorType pitype >
    typename Traits::template Codim< codim >::template Partition< pitype >::LeafIterator
    leafbegin ( const unsigned int sweepDir = 0 ) const
    {
      const LeafGridView &view = leafGridView();
      return getRealImplementation( view ).template begin< codim, pitype >( sweepDir );
    }

    template< int codim, PartitionIteratorType pitype >
    typename Traits::template Codim< codim >::template Partition< pitype >::LeafIterator
    leafend ( const unsigned int sweepDir = 0 ) const
    {
      const LeafGridView &view = leafGridView();
      return getRealImplementation( view ).template end< codim, pitype >( sweepDir );
    }

    template< int codim >
    typename Traits::template Codim< codim >::LeafIterator
    leafbegin ( const unsigned int sweepDir = 0 ) const
    {
      const LeafGridView &view = leafGridView();
      return getRealImplementation( view ).template begin< codim >( sweepDir );
    }

    template< int codim >
    typename Traits::template Codim< codim >::LeafIterator
    leafend ( const unsigned int sweepDir = 0 ) const
    {
      const LeafGridView &view = leafGridView();
      return getRealImplementation( view ).template end< codim >( sweepDir );
    }

    const GlobalIdSet &globalIdSet () const
    {
      return globalIdSet_;
    }

    const LocalIdSet &localIdSet () const
    {
      return localIdSet_;
    }

    const LevelIndexSet &levelIndexSet ( const int level ) const
    {
      return levelGridView( level ).indexSet();
    }

    const LeafIndexSet &leafIndexSet () const
    {
      return leafGridView().indexSet();
    }

    const HierarchicIndexSet &hierarchicIndexSet () const
    {
      return hierarchicIndexSet_;
    }

    bool mark ( const int refCount, const typename Codim< 0 >::Entity &e );
    int getMark ( const typename Codim< 0 >::Entity &e ) const;

    bool preAdapt ();

    bool adapt ();

    template< class DataHandle >
    bool adapt ( AdaptDataHandleInterface< This, DataHandle > &handle );

    void postAdapt ();

    void globalRefine ( const int refCount,
                        const RefinementPolicy &policy = RefinementPolicy() );

    template< class DataHandle >
    void globalRefine ( const int refCount,
                        AdaptDataHandleInterface< This, DataHandle > &handle,
                        const RefinementPolicy &policy = RefinementPolicy() );

    int overlapSize ( const int level, const int codim ) const
    {
      return levelGridView( level ).overlapSize( codim );
    }

    int overlapSize ( const int codim ) const
    {
      return leafGridView().overlapSize( codim );
    }

    int ghostSize ( const int level, const int codim ) const
    {
      return levelGridView( level ).ghostSize( codim );
    }

    int ghostSize ( const int codim ) const
    {
      return leafGridView().ghostSize( codim );
    }

    template< class DataHandle, class Data >
    void communicate ( CommDataHandleIF< DataHandle, Data > &data,
                       InterfaceType interface, CommunicationDirection dir,
                       int level ) const
    {
      levelGridView( level ).communicate( data, interface, dir );
    }

    template< class DataHandle, class Data >
    void communicate ( CommDataHandleIF< DataHandle, Data > &data,
                       InterfaceType interface, CommunicationDirection dir ) const
    {
      leafGridView().communicate( data, interface, dir );
    }

    const CollectiveCommunication &comm () const;

    template< class EntitySeed >
    typename Traits::template Codim< EntitySeed::codimension >::EntityPointer
    entityPointer ( const EntitySeed &seed ) const
    {
      typedef typename Traits::template Codim< EntitySeed::codimension >::EntityPointerImpl EntityPointerImpl;
      return EntityPointerImpl( gridLevel( getRealImplementation( seed ).level() ), getRealImplementation( seed ).id(), getRealImplementation( seed ).partitionNumber() );
    }

    const GridLevel &gridLevel ( const int level ) const;
    const GridLevel &leafLevel () const;

    size_t numBoundarySegments () const;

  private:
    // note: this method ignores the last bit of the macroId
    size_t boundaryIndex ( const MultiIndex &macroId,
                           const unsigned int partitionNumber,
                           const int face ) const;

    typename Codim< 1 >::LocalGeometry localFaceGeometry ( int face ) const
    {
      assert( (face >= 0) && (face < ReferenceCube::numFaces) );
      return typename Codim< 1 >::LocalGeometry( *localFaceGeometry_[ face ] );
    }

    void clear ();

    void createLocalGeometries ();
    void setupMacroGrid ();
    void setupBoundaryIndices ();

    static CollectiveCommunication defaultCommunication ();
  
    Domain domain_;
    Mesh globalMesh_;
    MultiIndex overlap_;
    ReferenceCubeContainer refCubes_;
    std::vector< GridLevel * > gridLevels_;
    std::vector< LevelGridView > levelGridViews_;
    LeafGridView leafGridView_;
    HierarchicIndexSet hierarchicIndexSet_;
    GlobalIdSet globalIdSet_;
    LocalIdSet localIdSet_;
    CollectiveCommunication comm_;
    size_t boundarySize_;
    std::vector< array< size_t, 2*dimension > > boundaryOffset_;
    const typename Codim< 1 >::LocalGeometryImpl *localFaceGeometry_[ ReferenceCube::numFaces ];
  };



  // Implementation of SPGrid
  // ------------------------

  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline SPGrid< ct, dim, strategy, Comm >::SPGrid ( const CollectiveCommunication &comm )
  : domain_( Domain::unitCube() ),
    globalMesh_( Mesh::unitMesh() ),
    overlap_( MultiIndex::zero() ),
    leafGridView_( LeafGridViewImpl() ),
    hierarchicIndexSet_( *this ),
    comm_( comm )
  {
    createLocalGeometries();
    setupMacroGrid();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline SPGrid< ct, dim, strategy, Comm >
    ::SPGrid ( const Domain &domain, const MultiIndex &cells,
               const CollectiveCommunication &comm )
  : domain_( domain ),
    globalMesh_( cells ),
    overlap_( MultiIndex::zero() ),
    leafGridView_( LeafGridViewImpl() ),
    hierarchicIndexSet_( *this ),
    comm_( comm )
  {
    createLocalGeometries();
    setupMacroGrid();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline SPGrid< ct, dim, strategy, Comm >
    ::SPGrid ( const Domain &domain, const MultiIndex &cells, const MultiIndex &overlap,
               const CollectiveCommunication &comm )
  : domain_( domain ),
    globalMesh_( cells ),
    overlap_( overlap ),
    leafGridView_( LeafGridViewImpl() ),
    hierarchicIndexSet_( *this ),
    comm_( comm )
  {
    createLocalGeometries();
    setupMacroGrid();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline SPGrid< ct, dim, strategy, Comm >
    ::SPGrid ( const GlobalVector &a, const GlobalVector &b, const MultiIndex &cells,
               const CollectiveCommunication &comm )
  : domain_( a, b ),
    globalMesh_( cells ),
    overlap_( MultiIndex::zero() ),
    leafGridView_( LeafGridViewImpl() ),
    hierarchicIndexSet_( *this ),
    comm_( comm )
  {
    createLocalGeometries();
    setupMacroGrid();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline SPGrid< ct, dim, strategy, Comm >
    ::SPGrid ( const GlobalVector &a, const GlobalVector &b, const MultiIndex &cells,
               const MultiIndex &overlap, const CollectiveCommunication &comm )
  : domain_( a, b ),
    globalMesh_( cells ),
    overlap_( overlap ),
    leafGridView_( LeafGridViewImpl() ),
    hierarchicIndexSet_( *this ),
    comm_( comm )
  {
    createLocalGeometries();
    setupMacroGrid();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  template< PartitionIteratorType pitype >
  typename SPGrid< ct, dim, strategy, Comm >::Traits::template Partition< pitype >::LevelGridView
  SPGrid< ct, dim, strategy, Comm >::levelGridView ( const int level ) const
  {
    typedef typename Traits::template Partition< pitype >::LevelGridView GridView;
    typedef typename GridView::Traits::GridViewImp GridViewImpl;
    assert( (level >= 0) && (level <= maxLevel()) );
    const LevelGridViewImpl &viewImpl = getRealImplementation( levelGridViews_[ level ] );
    return GridViewImpl( viewImpl );
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  template< PartitionIteratorType pitype >
  typename SPGrid< ct, dim, strategy, Comm >::Traits::template Partition< pitype >::LeafGridView
  SPGrid< ct, dim, strategy, Comm >::leafGridView () const
  {
    typedef typename Traits::template Partition< pitype >::LeafGridView GridView;
    typedef typename GridView::Traits::GridViewImp GridViewImpl;
    const LeafGridViewImpl &viewImpl = getRealImplementation( leafGridView_ );
    return GridViewImpl( viewImpl );
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  typename SPGrid< ct, dim, strategy, Comm >::LevelGridView
  SPGrid< ct, dim, strategy, Comm >::levelGridView ( const int level ) const
  {
    assert( (level >= 0) && (level <= maxLevel()) );
    return levelGridViews_[ level ];
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  typename SPGrid< ct, dim, strategy, Comm >::LeafGridView
  SPGrid< ct, dim, strategy, Comm >::leafGridView () const
  {
    return leafGridView_;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline bool SPGrid< ct, dim, strategy, Comm >
    ::mark ( const int refCount, const typename Codim< 0 >::Entity &e )
  {
    return false;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline int SPGrid< ct, dim, strategy, Comm >
    ::getMark ( const typename Codim< 0 >::Entity &e ) const
  {
    return 0;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline bool SPGrid< ct, dim, strategy, Comm >::preAdapt ()
  {
    return false;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline bool SPGrid< ct, dim, strategy, Comm >::adapt ()
  {
    return false;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  template< class DataHandle >
  inline bool SPGrid< ct, dim, strategy, Comm >
    ::adapt ( AdaptDataHandleInterface< This, DataHandle > &handle )
  {
    return false;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >::postAdapt ()
  {}


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >
    ::globalRefine ( const int refCount, const RefinementPolicy &policy )
  {
    for( int i = 0; i < refCount; ++i )
    {
      gridLevels_.push_back( new GridLevel( leafLevel(), policy ) );
      levelGridViews_.push_back( LevelGridViewImpl( leafLevel() ) );
    }
    getRealImplementation( leafGridView_ ).update( leafLevel() );
    hierarchicIndexSet_.update();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  template< class DataHandle >
  inline void SPGrid< ct, dim, strategy, Comm >
    ::globalRefine ( const int refCount,
                     AdaptDataHandleInterface< This, DataHandle > &handle,
                     const RefinementPolicy &policy )
  {
    for( int i = 0; i < refCount; ++i )
    {
      const LevelGridView fatherView = levelGridView( maxLevel() );

      gridLevels_.push_back( new GridLevel( leafLevel(), policy ) );
      levelGridViews_.push_back( LevelGridViewImpl( leafLevel() ) );

      hierarchicIndexSet_.update();
      getRealImplementation( leafGridView_ ).update( leafLevel() );

      handle.preAdapt( leafLevel().size() );
      typedef typename Codim< 0 >::LevelIterator LevelIterator;
      const LevelIterator end = fatherView.template end< 0 >();
      for( LevelIterator it = fatherView.template begin< 0 >(); it != end; ++it )
        handle.postRefinement( *it );
      handle.postAdapt();
    }
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline const typename SPGrid< ct, dim, strategy, Comm >::CollectiveCommunication &
  SPGrid< ct, dim, strategy, Comm >::comm () const
  {
    return comm_;
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline const typename SPGrid< ct, dim, strategy, Comm >::GridLevel &
  SPGrid< ct, dim, strategy, Comm >::gridLevel ( const int level ) const
  {
    assert( (level >= 0) && (level < int( gridLevels_.size() )) );
    return *gridLevels_[ level ];
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline const typename SPGrid< ct, dim, strategy, Comm >::GridLevel &
  SPGrid< ct, dim, strategy, Comm >::leafLevel () const
  {
    assert( !gridLevels_.empty() );
    return *gridLevels_.back();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline size_t SPGrid< ct, dim, strategy, Comm >::numBoundarySegments () const
  {
    return boundarySize_;
  }


  // note: this method ignores the last bit of the macroId
  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline size_t SPGrid< ct, dim, strategy, Comm >
    ::boundaryIndex ( const MultiIndex &macroId,
                      const unsigned int partitionNumber,
                      const int face ) const
  {
    assert( (face >= 0) && (face < 2*dimension) );

    const LevelGridView &macroView = levelGridView( 0 );
    const GridLevel &gridLevel = getRealImplementation( macroView ).gridLevel();
    const PartitionList &partitions = gridLevel.template partition< OverlapFront_Partition >();
    const typename PartitionList::Partition &partition = partitions.partition( partitionNumber );

    size_t index = 0;
    size_t factor = 1;
    for( int i = 0; i < dimension; ++i )
    {
      if( i == face/2 )
        continue;
      // note: the OverlapFront_Partition is closed, i.e.,
      //       begin()[ i ] & 1 == end()[ i ] & 1 == 0.
      const int k = (macroId[ i ] - partition.begin()[ i ]) >> 1;
      const int w = (partition.end()[ i ] - partition.begin()[ i ]) >> 1;
      assert( (k >= 0) && (k < w) );
      index += size_t( k ) * factor;
      factor *= size_t( w );
    }
    return index + boundaryOffset_[ partitionNumber - partitions.minNumber() ][ face ];
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >::clear ()
  {
    levelGridViews_.clear();
    leafGridView_ = LeafGridView( LeafGridViewImpl() );

    typedef typename std::vector< GridLevel * >::iterator Iterator;
    const Iterator end = gridLevels_.end();
    for( Iterator it = gridLevels_.begin(); it != end; ++it )
      delete *it;
    gridLevels_.clear();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >::createLocalGeometries ()
  {
    typedef typename Codim< 1 >::LocalGeometryImpl LocalGeometryImpl;

    const GlobalVector unitH( ctype( 1 ) );
    for( int face = 0; face < ReferenceCube::numFaces; ++face )
    {
      const unsigned int direction = ((1 << dimension) - 1) ^ (1 << (face/2));
      GlobalVector origin( ctype( 0 ) );
      origin[ face/2 ] = ctype( face & 1 );
      const SPGeometryCache< ctype, dimension, 1 > cache( unitH, direction );
      localFaceGeometry_[ face ] = new LocalGeometryImpl( cache, origin );
    }
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >::setupMacroGrid ()
  {
    clear();

    SPDecomposition< dimension > decomposition( globalMesh_, comm().size() );

    GridLevel *leafLevel = new GridLevel( *this, decomposition );
    gridLevels_.push_back( leafLevel );
    levelGridViews_.push_back( LevelGridViewImpl( *leafLevel ) );
    getRealImplementation( leafGridView_ ).update( *leafLevel );
    hierarchicIndexSet_.update();
    setupBoundaryIndices();
  }


  template< class ct, int dim, SPRefinementStrategy strategy, class Comm >
  inline void SPGrid< ct, dim, strategy, Comm >::setupBoundaryIndices ()
  {
    const LevelGridView &macroView = levelGridView( 0 );
    const GridLevel &gridLevel = getRealImplementation( macroView ).gridLevel();
    const PartitionList &partitions = gridLevel.template partition< OverlapFront_Partition >();

    boundarySize_ = 0;
    boundaryOffset_.resize( partitions.maxNumber() - partitions.minNumber() + 1 );
    for( typename PartitionList::Iterator it = partitions.begin(); it; ++it )
    {
      const int partitionIndex = it->number() - partitions.minNumber();
      for( int i = 0; i < dimension; ++i )
      {
        // note: the OverlapFront_Partition is closed, i.e.,
        //       begin()[ i ] & 1 == end()[ i ] & 1 == 0.
        size_t size = 1;
        for( int j = 0; j < dimension; ++j )
          size *= (i == j ? 1 : size_t( (it->end()[ j ] - it->begin()[ j ]) >> 1 ));

        // offset for lower boundary
        boundaryOffset_[ partitionIndex ][ 2*i ] = boundarySize_;
        boundarySize_ += (it->boundary( 2*i ) ? size : 0);

        // offset for upper boundary
        boundaryOffset_[ partitionIndex ][ 2*i+1 ] = boundarySize_;
        boundarySize_ += (it->boundary( 2*i+1 ) ? size : 0);
      }
    }
  }

} // namespace Dune

#endif // #ifndef DUNE_SPGRID_GRID_HH
